from llmcast import TokenUsage


def test_defaults_to_zero():
    usage = TokenUsage()
    assert usage.prompt_tokens == 0
    assert usage.completion_tokens == 0
    assert usage.total_tokens == 0


def test_addition():
    a = TokenUsage(prompt_tokens=10, completion_tokens=5, total_tokens=15)
    b = TokenUsage(prompt_tokens=20, completion_tokens=10, total_tokens=30)
    c = a + b
    assert c.prompt_tokens == 30
    assert c.completion_tokens == 15
    assert c.total_tokens == 45


def test_addition_with_zero():
    a = TokenUsage(prompt_tokens=5, completion_tokens=3, total_tokens=8)
    b = TokenUsage()
    assert (a + b) == a


def test_addition_is_not_mutating():
    a = TokenUsage(prompt_tokens=10, completion_tokens=5, total_tokens=15)
    b = TokenUsage(prompt_tokens=1, completion_tokens=1, total_tokens=2)
    _ = a + b
    assert a.prompt_tokens == 10



from jinja2 import Environment
from pydantic import BaseModel
from textwrap import dedent
from typing import Literal

type ResultFormat = Literal["json", "yaml", "toml"]

env = Environment()


class BaseTemplate(BaseModel):
    output_format: ResultFormat = "json"

    @property
    def template(self) -> str:
        if doc := self.__doc__:
            return dedent(doc)
        return ""

    def __str__(self) -> str:
        template_string = self.template

        if template_string is None:
            raise ValueError("Template string (docstring) not found")

        template = env.from_string(template_string)
        values = self.model_dump(mode="json")

        return template.render(values)



import random
import tomllib
from dataclasses import dataclass
from typing import Type

import yaml
from loguru import logger
from pydantic import BaseModel, ValidationError

from llmcast.template import BaseTemplate, ResultFormat


@dataclass
class RetryPolicy:
    n_tries: int = 1
    backoff: float = 1.0  # initial delay in seconds
    multiplier: float = 2.0  # exponential factor per attempt
    max_backoff: float = 60.0  # upper bound on delay
    jitter: bool = True  # randomize delay to avoid thundering herd

    def delay_for(self, attempt: int) -> float:
        """Return sleep duration (seconds) before the given attempt (0-indexed)."""
        if attempt == 0:
            return 0.0
        delay = min(self.backoff * (self.multiplier ** (attempt - 1)), self.max_backoff)
        if self.jitter:
            delay *= 0.5 + random.random() * 0.5
        return delay


@dataclass
class TokenUsage:
    prompt_tokens: int = 0
    completion_tokens: int = 0
    total_tokens: int = 0

    def __add__(self, other: "TokenUsage") -> "TokenUsage":
        return TokenUsage(
            prompt_tokens=self.prompt_tokens + other.prompt_tokens,
            completion_tokens=self.completion_tokens + other.completion_tokens,
            total_tokens=self.total_tokens + other.total_tokens,
        )


def build_messages(prompt: BaseTemplate, query: str | None) -> list[dict]:
    messages = [{"role": "system", "content": str(prompt)}]
    if query:
        messages.append({"role": "user", "content": query})
    return messages


def parse_content[TResult: BaseModel](
    content: str,
    result_schema: Type[TResult],
    result_format: ResultFormat = "json",
) -> TResult | None:
    content = content.replace(f"```{result_format}", "").replace("```", "").strip()
    try:
        match result_format:
            case "json":
                return result_schema.model_validate_json(content)
            case "yaml":
                return result_schema.model_validate(yaml.safe_load(content))
            case "toml":
                return result_schema.model_validate(tomllib.loads(content))
    except (ValidationError, yaml.YAMLError, tomllib.TOMLDecodeError) as e:
        logger.error(e)
        logger.debug(content)
        return None



from pydantic import BaseModel
from llmcast.parser.utils import parse_content


class Result(BaseModel):
    value: int
    name: str


def test_parse_json():
    result = parse_content('{"value": 42, "name": "test"}', Result)
    assert result is not None
    assert result.value == 42
    assert result.name == "test"


def test_parse_json_strips_fences():
    result = parse_content('```json\n{"value": 1, "name": "a"}\n```', Result)
    assert result is not None
    assert result.value == 1


def test_parse_json_strips_plain_fences():
    result = parse_content('```\n{"value": 2, "name": "b"}\n```', Result)
    assert result is not None
    assert result.value == 2


def test_parse_yaml():
    result = parse_content("value: 10\nname: hello", Result, result_format="yaml")
    assert result is not None
    assert result.value == 10
    assert result.name == "hello"


def test_parse_yaml_strips_fences():
    result = parse_content(
        "```yaml\nvalue: 5\nname: y\n```", Result, result_format="yaml"
    )
    assert result is not None
    assert result.value == 5


def test_parse_toml():
    result = parse_content('value = 7\nname = "world"', Result, result_format="toml")
    assert result is not None
    assert result.value == 7
    assert result.name == "world"


def test_invalid_json_returns_none():
    result = parse_content("not valid json", Result)
    assert result is None


def test_invalid_yaml_returns_none():
    result = parse_content("value: [unclosed", Result, result_format="yaml")
    assert result is None


def test_schema_mismatch_returns_none():
    result = parse_content('{"value": "not_an_int", "name": "x"}', Result)
    assert result is None



from llmcast import BaseTemplate


class GreetTemplate(BaseTemplate):
    """Hello, {{ name }}! Format: {{ output_format }}."""

    name: str


class EmptyTemplate(BaseTemplate):
    pass


def test_renders_variables():
    tmpl = GreetTemplate(name="World")
    assert str(tmpl) == "Hello, World! Format: json."


def test_default_output_format():
    tmpl = GreetTemplate(name="x")
    assert tmpl.output_format == "json"


def test_custom_output_format():
    tmpl = GreetTemplate(name="x", output_format="yaml")
    assert tmpl.output_format == "yaml"
    assert "yaml" in str(tmpl)


def test_empty_docstring_returns_empty():
    tmpl = EmptyTemplate()
    assert str(tmpl) == ""


def test_jinja2_conditional():
    class ConditionalTemplate(BaseTemplate):
        """{% if flag %}yes{% else %}no{% endif %}"""

        flag: bool

    assert str(ConditionalTemplate(flag=True)) == "yes"
    assert str(ConditionalTemplate(flag=False)) == "no"


def test_jinja2_loop():
    class ListTemplate(BaseTemplate):
        """{% for item in items %}{{ item }}{% endfor %}"""

        items: list[str]

    assert str(ListTemplate(items=["a", "b", "c"])) == "abc"



import asyncio
from typing import Type

import openai
from openai import AsyncOpenAI
from pydantic import BaseModel
from loguru import logger

from llmcast.parser.utils import RetryPolicy, TokenUsage, build_messages, parse_content
from llmcast.template import BaseTemplate

_RETRYABLE = (openai.RateLimitError, openai.APITimeoutError, openai.InternalServerError)


class AsyncLLMParser:
    def __init__(
        self,
        client: AsyncOpenAI,
        model_name: str,
        timeout: int = 10,
        concurrency_limit: int = 10,
        retry_policy: RetryPolicy | None = None,
        structured_output: bool = True,
    ):
        self._client = client
        self._model_name = model_name
        self._timeout = timeout
        self._retry_policy = retry_policy or RetryPolicy()
        self._structured_output = structured_output
        self._semaphore = asyncio.Semaphore(concurrency_limit)

    async def _complete(self, messages: list[dict]) -> tuple[str | None, TokenUsage]:
        async with self._semaphore:
            completion = await self._client.chat.completions.create(
                model=self._model_name,
                messages=messages,  # type: ignore
                timeout=self._timeout,
            )
        usage = TokenUsage(
            prompt_tokens=completion.usage.prompt_tokens if completion.usage else 0,
            completion_tokens=(
                completion.usage.completion_tokens if completion.usage else 0
            ),
            total_tokens=completion.usage.total_tokens if completion.usage else 0,
        )
        return completion.choices[0].message.content, usage

    async def _parse[TResult: BaseModel](
        self, messages: list[dict], result_schema: Type[TResult]
    ) -> tuple[TResult | None, TokenUsage]:
        async with self._semaphore:
            completion = await self._client.chat.completions.parse(
                model=self._model_name,
                messages=messages,  # type: ignore
                response_format=result_schema,
                timeout=self._timeout,
            )
        usage = TokenUsage(
            prompt_tokens=completion.usage.prompt_tokens if completion.usage else 0,
            completion_tokens=(
                completion.usage.completion_tokens if completion.usage else 0
            ),
            total_tokens=completion.usage.total_tokens if completion.usage else 0,
        )
        return completion.choices[0].message.parsed, usage

    async def parse[TResult: BaseModel](
        self,
        prompt: BaseTemplate,
        result_schema: Type[TResult],
        query: str | None = None,
        retry_policy: RetryPolicy | None = None,
    ) -> tuple[TResult, TokenUsage] | None:
        if self._structured_output:
            assert prompt.output_format == "json", (
                "Structured output only supported for JSON"
            )

        messages = build_messages(prompt, query)
        policy = retry_policy or self._retry_policy
        total_usage = TokenUsage()

        for attempt in range(policy.n_tries):
            if (delay := policy.delay_for(attempt)) > 0:
                await asyncio.sleep(delay)
            try:
                if self._structured_output:
                    result, usage = await self._parse(messages, result_schema)
                    total_usage = total_usage + usage
                    if result is not None:
                        return result, total_usage
                else:
                    content, usage = await self._complete(messages)
                    total_usage = total_usage + usage
                    if content and (
                        result := parse_content(
                            content, result_schema, result_format=prompt.output_format
                        )
                    ):
                        return result, total_usage
            except _RETRYABLE as e:
                if attempt == policy.n_tries - 1:
                    raise
                logger.warning(
                    "Retryable API error (attempt {}/{}): {}",
                    attempt + 1,
                    policy.n_tries,
                    e,
                )
                continue

            logger.debug(
                "Parse failed on attempt {}/{}, retrying", attempt + 1, policy.n_tries
            )

        return None



from llmcast.template import BaseTemplate, ResultFormat
from llmcast.parser.utils import RetryPolicy, TokenUsage
from llmcast.parser.sync import SyncLLMParser
from llmcast.parser.async_ import AsyncLLMParser

__all__ = [
    "BaseTemplate",
    "ResultFormat",
    "RetryPolicy",
    "TokenUsage",
    "SyncLLMParser",
    "AsyncLLMParser",
]



from llmcast import RetryPolicy


def test_first_attempt_no_delay():
    policy = RetryPolicy()
    assert policy.delay_for(0) == 0.0


def test_exponential_backoff():
    policy = RetryPolicy(backoff=1.0, multiplier=2.0, jitter=False)
    assert policy.delay_for(1) == 1.0
    assert policy.delay_for(2) == 2.0
    assert policy.delay_for(3) == 4.0


def test_max_backoff_capped():
    policy = RetryPolicy(backoff=1.0, multiplier=2.0, max_backoff=3.0, jitter=False)
    assert policy.delay_for(5) == 3.0


def test_jitter_within_range():
    policy = RetryPolicy(backoff=1.0, multiplier=2.0, jitter=True)
    for _ in range(20):
        delay = policy.delay_for(1)
        assert 0.5 <= delay <= 1.0


def test_default_single_try():
    policy = RetryPolicy()
    assert policy.n_tries == 1
